"""UK Data sources."""

CONST_PETROLPRICES_BASE = "https://mobile.app.petrolprices.com/api"
CONST_PETROLPRICES_LOGIN = "{BASE}/guest-mode/gello"
CONST_PETROLPRICES_FUELSTATIONS = "{BASE}/petrolstationsgeo/{FUEL_TYPE}/0/50/0/{RADIUS}/price?lat={LAT}&lng={LNG}"
CONST_PETROLPRICES_FUEL_MAP = {
    "1": "E5",
    "2": "E10",
    "4": "B10",
    "5": "B7"
}

CONST_PODPOINT_BASE = "https://mobile-api.pod-point.com/api3/v5"
CONST_PODPOINT_SEARCH = "{BASE}/addresses?payg=true&force=true&order=proximity:{LAT}:{LNG}"
CONST_PODPOINT_LOCATION = "{BASE}/addresses/{LOC_ID}/pods?include=price,model,charges,advert,unit_connectors.status,range&perpage=all&order=-availability"



"""PodPoint data source."""

import asyncio
import logging

from datetime import datetime

from pyfuelprices.const import (
    PROP_AREA_LAT,
    PROP_AREA_LONG
)
from pyfuelprices.sources import Source, geocode_reverse_lookup, geopyexc
from pyfuelprices.fuel_locations import Fuel, FuelLocation

from .const import CONST_PODPOINT_BASE, CONST_PODPOINT_LOCATION, CONST_PODPOINT_SEARCH

_LOGGER = logging.getLogger(__name__)

class PodPointSource(Source):
    """Pod Point data source."""
    provider_name="podpoint"
    _headers = {
        "User-Agent": "POD Point Native Mobile App/3.27.12 (Android/14)"
    }
    location_cache: dict[str, FuelLocation] = {}

    async def _send_request(self, url) -> dict:
        """Send a request to the API and return the raw response."""
        _LOGGER.debug("Sending HTTP request to FuelGR with URL %s", url)
        async with self._client_session.get(url=url,
                                            headers=self._headers) as response:
            if response.ok:
                return await response.json()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites (radius not used)"""
        # first query the API to populate the cache / update data
        data = await super().search_sites(coordinates, radius)
        if len(data)>0:
            return data
        await self.update_area(
            {
                PROP_AREA_LAT: coordinates[0],
                PROP_AREA_LONG: coordinates[1]
            }
        )
        return await super().search_sites(coordinates, radius)

    async def update_area(self, area: dict):
        """Update a single area."""
        try:
            geocode = await geocode_reverse_lookup(
                (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
            )
        except geopyexc.GeocoderTimedOut:
            _LOGGER.warning("Timeout occured while geocoding area %s.",
                            area)
            return None
        if geocode.raw["address"]["country_code"] != "gb":
            _LOGGER.debug("Skipping area %s as not in GB.",
                        area)
            return None
        _LOGGER.debug("Searching Pod Point for FuelLocations at area %s", area)
        response_raw = await self._send_request(
            url=CONST_PODPOINT_SEARCH.format(
                BASE=CONST_PODPOINT_BASE,
                LAT=area[PROP_AREA_LAT],
                LNG=area[PROP_AREA_LONG]
            )
        )
        await self.parse_response(response_raw)

    async def get_and_parse_pod(self, response: dict):
        """Retrieves pod details and parses the response."""
        _LOGGER.debug("Parsing object %s", response["id"])
        site_id = f"{self.provider_name}_{response['id']}"
        fs = FuelLocation.create(
            site_id=site_id,
            name=response["name"],
            address="See properties",
            lat=response["location"]["lat"],
            long=response["location"]["lng"],
            brand="Pod Point",
            available_fuels=[],
            currency="GBP",
            postal_code=response["address"]["postcode"],
            next_update=self.next_update,
            props=response
        )
        pods = self.parse_fuels(await self._send_request(
            url=CONST_PODPOINT_LOCATION.format(
                BASE=CONST_PODPOINT_BASE,
                LOC_ID=response["id"]
            )
        ))
        for pod in pods:
            fs.add_or_update_fuel(pod)
            fs.next_update=self.next_update
            fs.last_updated=datetime.now()
        if site_id not in self.location_cache:
            self.location_cache[site_id] = fs
        else:
            self.location_cache.get(site_id).update(fs)
        _LOGGER.debug("Parsed object %s", response["id"])

    def parse_fuels(self, fuels: list[dict]):
        """Parse fuels."""
        # EV will use the fuel type format "EV-POWER kW"
        fuels_parsed = []
        fuels = fuels["pods"]
        for f in fuels:
            if "unit_connectors" not in f:
                continue
            if not isinstance(f["unit_connectors"], list):
                continue
            if len(f["unit_connectors"]) == 0:
                continue
            if "price" not in f:
                continue
            if "cost" not in f["price"]:
                continue
            if not isinstance(f["price"]["cost"], list):
                continue
            if len(f["price"]["cost"]) == 0:
                continue
            connector = f["unit_connectors"][0]['connector']
            fuels_parsed.append(Fuel(
                fuel_type=f"EV-{connector['power']} kW",
                cost=f["price"]["cost"][0]["price"]/100,
                props=f
            ))
        return fuels_parsed

    async def parse_response(self, response):
        """Parse fuel station response."""
        coros = [self.get_and_parse_pod(x) for x in response["addresses"]]
        await asyncio.gather(*coros)

    async def update(self, areas=None, force=False):
        """Update data source."""
        if self.next_update <= datetime.now() or force:
            self._configured_areas=[] if areas is None else areas
            try:
                self.next_update += self.update_interval
                coros = [self.update_area(a) for a in self._configured_areas]
                await asyncio.gather(*coros)
            except Exception as exc:
                _LOGGER.exception(exc, exc_info=exc)

        return list(self.location_cache.values())



"""Sources mapping file."""

from .argentina import SOURCE_MAP as AR_SOURCE_MAP
from .austria import SOURCE_MAP as AT_SOURCE_MAP
from .australia import SOURCE_MAP as AU_SOURCE_MAP
from .belgium import SOURCE_MAP as BE_SOURCE_MAP
from .brazil import SOURCE_MAP as BR_SOURCE_MAP
from .canada import SOURCE_MAP as CA_SOURCE_MAP
from .germany import SOURCE_MAP as DE_SOURCE_MAP
from .greece import SOURCE_MAP as GR_SOURCE_MAP
from .romania import SOURCE_MAP as RO_SOURCE_MAP
from .uk.map import SOURCE_MAP as UK_SOURCE_MAP
from .usa import SOURCE_MAP as US_SOURCE_MAP
from .netherlands import SOURCE_MAP as NL_SOURCE_MAP
from .new_zealand import SOURCE_MAP as NZ_SOURCE_MAP
from .slovenia import SOURCE_MAP as SI_SOURCE_MAP
from .switzerland import SOURCE_MAP as CH_SOURCE_MAP

# Fields are as follows:
# Key - Provider Name
# Value - (Provider Class, Enabled, Available, Country Mapping Enabled)
# Available field is used to control what options are
# shown in the Home Assistant config wizard.
# The last int is an optional value that defines if the provider will be
# included in the area auto mapping system, if set to 0, the provider
# can only be configured manually

SOURCE_MAP: dict[str, tuple[object, int, int]] = {
    **UK_SOURCE_MAP,
    **NL_SOURCE_MAP,
    **US_SOURCE_MAP,
    **DE_SOURCE_MAP,
    **AT_SOURCE_MAP,
    **AU_SOURCE_MAP,
    **CH_SOURCE_MAP,
    **RO_SOURCE_MAP,
    **GR_SOURCE_MAP,
    **BR_SOURCE_MAP,
    **AR_SOURCE_MAP,
    **SI_SOURCE_MAP,
    **BE_SOURCE_MAP,
    **CA_SOURCE_MAP,
    **NZ_SOURCE_MAP
}

FULL_COUNTRY_MAP = {
    "AR": [k for k in AR_SOURCE_MAP],
    "BE": [k for k in BE_SOURCE_MAP],
    "BR": [k for k in BR_SOURCE_MAP],
    "CA": [k for k in CA_SOURCE_MAP],
    "DE": [k for k in DE_SOURCE_MAP],
    "NL": [k for k in NL_SOURCE_MAP],
    "NZ": [k for k in NZ_SOURCE_MAP],
    "GB": [k for k in UK_SOURCE_MAP],
    "US": [k for k in US_SOURCE_MAP],
    "AU": [k for k in AU_SOURCE_MAP],
    "AT": [k for k in AT_SOURCE_MAP],
    "CH": [k for k in CH_SOURCE_MAP],
    "RO": [k for k in RO_SOURCE_MAP],
    "GR": [k for k in GR_SOURCE_MAP],
    "SI": [k for k in SI_SOURCE_MAP]
}

COUNTRY_MAP = {
    "AR": [k for k, v in AR_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "BE": [k for k, v in BE_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "BR": [k for k, v in BR_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "CA": [k for k, v in CA_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "DE": [k for k, v in DE_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "NL": [k for k, v in NL_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "NZ": [k for k, v in NZ_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "GB": [k for k, v in UK_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "US": [k for k, v in US_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "AU": [k for k, v in AU_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "AT": [k for k, v in AT_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "CH": [k for k, v in CH_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "RO": [k for k, v in RO_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "GR": [k for k, v in GR_SOURCE_MAP.items() if len(v)==3 and v[1]==1],
    "SI": [k for k, v in SI_SOURCE_MAP.items() if len(v)==3 and v[1]==1]
}



"""Hold source mappings for modules."""

from pyfuelprices.sources import Source
from .applegreenstores import ApplegreenUKSource
from .ascona import AsconaGroupUKSource
from .asda import AsdaUKSource
from .bpuk import BpUKSource
from .costco import CostcoUKSource
from .essouk import EssoUKSource
from .jet import JetUKSource
from .karanretail import KaranRetailSource
from .moto import MotowayUKSource
from .motorfuelgroup import MotorFuelGroupUKSource
from .petrolprices import PetrolPricesUKSource
from .podpoint import PodPointSource
from .rontec import RontecUKSource
from .sainsburys import SainsburysUKSource
from .sgn import SgnRetailUKSource
from .shell import ShellUKSource
from .tesco import TescoUKSource

SOURCE_MAP: dict[str, tuple[Source, int, int]] = {
    "applegreen": (ApplegreenUKSource, 1, 1,),
    "asda": (AsdaUKSource, 1, 1,),
    "ascona": (AsconaGroupUKSource, 1, 1,),
    "bpuk": (BpUKSource, 1, 1,),
    "costco": (CostcoUKSource, 1, 1,),
    "essouk": (EssoUKSource, 1, 1,),
    "jet": (JetUKSource, 1, 0,),
    "karanretail": (KaranRetailSource, 1, 1,),
    "motoway": (MotowayUKSource, 1, 1,),
    "motorfuelgroup": (MotorFuelGroupUKSource, 1, 1,),
    "petrolprices": (PetrolPricesUKSource, 1, 1, 1),
    "podpoint": (PodPointSource, 1, 1, 0,),
    "rontec": (RontecUKSource, 1, 1,),
    "sainsburys": (SainsburysUKSource, 1, 1,),
    "sgnretail": (SgnRetailUKSource, 1, 1,),
    "shelluk": (ShellUKSource, 1, 1,),
    "tesco": (TescoUKSource, 1, 1,)
}



"""Australia data sources."""

from .fuelwatch import FuelWatchSource
from .fuelsnoop import FuelSnoopSource
from .petrolspy import PetrolSpySource

SOURCE_MAP = {
    "fuelwatch": (FuelWatchSource, 1, 1,),
    "fuelsnoop": (FuelSnoopSource, 1, 1,),
    "petrolspy": (PetrolSpySource, 1, 1,)
}



"""Costco Data Source."""

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID,
    DESKTOP_USER_AGENT
)
from pyfuelprices.fuel_locations import Fuel, FuelLocation
from pyfuelprices.sources import Source

COSTCO_FUEL_MAPPING = {
    "Premium Unleaded Petrol": "E5",
    "Premium Diesel": "B7",
    "Unleaded Petrol": "E10"
}

class CostcoUKSource(Source):
    """A Costco data source."""
    _url = "https://www.costco.co.uk/store-finder/search?q=United%20Kingdom&page=0"
    _headers = {
        "User-Agent": DESKTOP_USER_AGENT,
        "Accept": "application/json",
        "Accept-Language": "en-GB,en;q=0.5",
        "X-Requested-With": "XMLHttpRequest",
        "Referer": "https://www.costco.co.uk/store-finder"
    }
    location_cache: dict[str, FuelLocation] = {}
    provider_name: str = "costco"

    def parse_fuels(self, fuels):
        """Parse Costco Fuel."""
        parsed = []
        for service in fuels:
            if service["code"] == "GAS_STATION":
                fuels = service["gasTypes"]
                break
        for fuel in fuels:
            parsed.append(Fuel(
                fuel_type=COSTCO_FUEL_MAPPING.get(fuel["name"], fuel["name"]),
                cost=float(fuel["price"])/100,
                props={}
            ))
        return parsed

    async def parse_response(self, response) -> list[FuelLocation]:
        if "data" in response:
            response = response["data"]
        for station in response:
            if "Petrol Station" not in station["features"]:
                continue
            site_id = f"{self.provider_name}_{station['addressId']}"
            location = FuelLocation.create(
                site_id=site_id,
                name=f"Costco {station["displayName"]}",
                address=f"{station["line1"]}\n{station["line2"]}\n{station["town"]}\n{station["postalCode"]}",
                brand=self.provider_name,
                available_fuels=self.parse_fuels(
                    station["availableServices"]
                ),
                postal_code=station["postalCode"],
                lat=station["latitude"],
                long=station["longitude"],
                currency="GBP",
                props={
                    PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                    PROP_FUEL_LOCATION_SOURCE_ID: station['addressId'],
                    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True
                },
                next_update=self.next_update
            )
            if site_id not in self.location_cache:
                self.location_cache[site_id] = location
            else:
                await self.location_cache[site_id].update(location)
        return list(self.location_cache.values())



"""Goriva Source."""

import logging
import json
from datetime import datetime

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID,
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS
)
from pyfuelprices.sources import Source, geocode_reverse_lookup, geopyexc
from pyfuelprices.fuel import Fuel
from pyfuelprices.fuel_locations import FuelLocation

_LOGGER = logging.getLogger(__name__)

class GorivaSource(Source):
    """Goriva fuel source."""

    provider_name="goriva"
    location_cache: dict[str, FuelLocation] = {}
    _url: str = "https://goriva.si/api/v1/search/?position={LAT},{LONG}&radius={RADIUS}&franchise=&name=&o="

    async def _send_request(self, url: str, lat, long, radius) -> str:
        """Send a request to the API and return the raw response."""
        url = url.format(
            LAT=lat,
            LONG=long,
            RADIUS=radius
        )
        _LOGGER.debug("Sending HTTP request to %s with URL %s", self.provider_name, url)
        radius = (radius * 1.609344)*1000 # convert miles to m
        async with self._client_session.get(url=url) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites within a given radius."""
        # first query the API to populate cache / update data in case this data is unavailable.
        data = await super().search_sites(coordinates, radius)
        if len(data)>0:
            return data
        await self.update(
            areas=[{
                PROP_AREA_LAT: coordinates[0],
                PROP_AREA_LONG: coordinates[1],
                PROP_AREA_RADIUS: radius
            }],
            force=True
        )
        data = await super().search_sites(coordinates, radius)
        return data

    async def update(self, areas=None, force=False) -> list[FuelLocation]:
        """Update hooks for the data source."""
        _LOGGER.debug("Starting update hook for %s to url %s", self.provider_name, self._url)
        self._configured_areas=[] if areas is None else areas
        self._clear_cache()
        if self.next_update <= datetime.now() or force:
            for area in self._configured_areas:
                try:
                    geocode = await geocode_reverse_lookup(
                        (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
                    )
                except geopyexc.GeocoderTimedOut:
                    _LOGGER.warning("Timeout occured while geocoding area %s.",
                                    area)
                    continue
                if geocode.raw["address"]["country_code"] != "si":
                    _LOGGER.debug("Skipping area %s as not in SI.",
                                area)
                    continue
                response = json.loads(await self._send_request(
                    url=self._url,
                    radius=area[PROP_AREA_RADIUS],
                    lat=area[PROP_AREA_LAT],
                    long=area[PROP_AREA_LONG]
                ))
                await self.parse_response(response["results"])
            self.next_update = datetime.now() + self.update_interval
        return list(self.location_cache.values())

    async def parse_response(self, response) -> list[FuelLocation]:
        """Converts CMA data into fuel price mapping."""
        for location_raw in response:
            site_id = f"{self.provider_name}_{location_raw['pk']}"
            location = FuelLocation.create(
                    site_id=site_id,
                    name=location_raw['name'],
                    address=location_raw["address"],
                    brand="",
                    lat=location_raw["lat"],
                    long=location_raw["lng"],
                    available_fuels=self.parse_fuels(location_raw["prices"]),
                    postal_code=location_raw["zip_code"],
                    currency="EUR",
                    props={
                        PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                        PROP_FUEL_LOCATION_SOURCE_ID: location_raw["pk"],
                        PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                        "open_hours": location_raw["open_hours"]
                    },
                    next_update=self.next_update
                )
            if site_id not in self.location_cache:
                self.location_cache[site_id] = location
            else:
                await self.location_cache[site_id].update(location)
        return list(self.location_cache.values())

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Parses fuel data."""
        output = []
        for k in fuels:
            if fuels[k] is None:
                continue
            output.append(
                Fuel(
                    fuel_type=k,
                    cost=fuels[k],
                    props={}
                )
            )
        return output



"""Data source for PetrolPrices UK."""

import asyncio
import logging

from datetime import datetime, timedelta

from pyfuelprices.const import (
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID
)

from pyfuelprices.sources import Source, FuelLocation, Fuel

from .const import (
    CONST_PETROLPRICES_BASE,
    CONST_PETROLPRICES_FUELSTATIONS,
    CONST_PETROLPRICES_LOGIN,
    CONST_PETROLPRICES_FUEL_MAP
)

_LOGGER = logging.getLogger(__name__)

class PetrolPricesUKSource(Source):
    """PetrolPrices source."""

    _headers = {
        "Accept": "application/json",
        "User-Agent": "okhttp/4.9.2"
    }
    _access_token: str = ""
    _refresh_token: str = ""
    _at_expires: datetime | None = None
    _at_update_lock: asyncio.Lock = asyncio.Lock()
    _fs_update_lock: asyncio.Lock = asyncio.Lock()
    provider_name = "petrolprices"
    location_cache = {}

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites within a given radius."""
        # first query the API to populate cache / update data in case this data is unavailable.
        data = await super().search_sites(coordinates, radius)
        if len(data)>0:
            return data
        await self.update(
            areas=[
                {
                    PROP_AREA_LAT: coordinates[0],
                    PROP_AREA_LONG: coordinates[1],
                    PROP_AREA_RADIUS: radius
                }
            ],
            force=True
        )
        return await super().search_sites(coordinates, radius)

    async def _send_request(self, url, method) -> None | dict:
        """Send a HTTP request."""
        _LOGGER.debug("Sending %s request to %s", method, url)
        if "guest-mode/gello" not in url:
            await self._refresh_access_token()
        else:
            self._access_token = ""
            self._refresh_token = ""
        async with self._client_session.request(
            method=method,
            url=url,
            timeout=30,
            headers={
                **self._headers,
                "authorization": self._access_token,
                "refresh-token": self._refresh_token
            }
        ) as resp:
            _LOGGER.debug("Got status code %s for request %s",
                          resp.status, url)
            if resp.status != 200:
                _LOGGER.error("Request to %s failed %s (%s)",
                              url,
                              resp.status,
                              await resp.text())
                return
            return await resp.json()

    async def _refresh_access_token(self):
        """Refresh the access token if needed."""
        async with self._at_update_lock:
            if self._at_expires is None or self._at_expires < datetime.now():
                _LOGGER.debug("Refreshing access token.")
                response = await self._send_request(
                    CONST_PETROLPRICES_LOGIN.format(BASE=CONST_PETROLPRICES_BASE),
                    "POST"
                )
                if response is None:
                    pass
                self._access_token = response["accessToken"]
                self._refresh_token = response["refreshToken"]
                self._at_expires = datetime.now() + timedelta(hours=23)

    async def update_area(self, area: dict, fuel_type: str, fuel_code: str):
        """Update a given area."""
        response = await self._send_request(
            CONST_PETROLPRICES_FUELSTATIONS.format(
                BASE=CONST_PETROLPRICES_BASE,
                FUEL_TYPE=fuel_type,
                RADIUS=area[PROP_AREA_RADIUS],
                LAT=area[PROP_AREA_LAT],
                LNG=area[PROP_AREA_LONG]
            ),
            method="GET"
        )
        if response is None:
            return
        response["fuel_code"] = fuel_code
        await self.parse_response(response)

    async def update(self, areas=None, force=False):
        """Update PetrolPrices data."""
        self._configured_areas = areas or []
        self._clear_cache()
        if self.next_update > datetime.now() and not force:
            return
        coros = [
            self.update_area(a, ft, code) for a in self._configured_areas
            for ft, code in CONST_PETROLPRICES_FUEL_MAP.items()
        ]
        results = await asyncio.gather(*coros)
        for result in results:
            if isinstance(result, Exception):
                _LOGGER.exception("Error updating area: %s", result)
        return list(self.location_cache.values())

    async def parse_or_update_station(self, station: dict, fuel_type: str):
        """Parse or update the fuel station."""
        _LOGGER.debug("Parsing station %s", station["idstation"])
        site_id = f"{self.provider_name}_{station['idstation']}"
        price = float(station["price"]) / 1000
        fuel = Fuel(
            fuel_type=fuel_type,
            cost=price if price > 0 else -1,
            props={
                **station["fuel"],
                "available": price>0}
        )
        loc = FuelLocation.create(
            site_id=site_id,
            name=station["name"],
            address="See properties",
            lat=station["lat"],
            long=station["lng"],
            brand=station["fuel_brand_name"],
            available_fuels=[] if fuel.cost == -1 else [fuel],
            postal_code=station["postcode"],
            currency="GBP",
            props={
                "data": station,
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station['idstation']
            },
            next_update=self.next_update+self.update_interval
        )
        async with self._fs_update_lock:
            if site_id not in self.location_cache:
                self.location_cache[site_id] = loc
            else:
                await self.location_cache[site_id].update(loc)
        return self.location_cache[site_id]

    async def parse_response(self, response: dict) -> list[FuelLocation]:
        """Parse a fuel response."""
        if "error" in response and response["error"]:
            _LOGGER.error("PetrolPrices API Returned an unknown error.")
            return
        if "data" not in response and not isinstance(response["data"], list):
            return
        for station in response["data"]:
            await self.parse_or_update_station(station, response["fuel_code"])
        return self.location_cache

    def parse_fuels(self, fuels):
        """Not used."""



"""GOB Energy Source."""

import logging
import json
from datetime import datetime, timedelta

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID,
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS
)
from pyfuelprices.sources import Source, geocode_reverse_lookup, geopyexc
from pyfuelprices.fuel import Fuel
from pyfuelprices.fuel_locations import FuelLocation

from .const import AR_GOB_ENERGY_ID, AR_GOB_DATASOURCE, AR_GOB_TIMEOUT

_LOGGER = logging.getLogger(__name__)

class GobEnergySource(Source):
    """GobEnergy fuel source."""

    update_interval = timedelta(hours=12) # hardcoded due to slow servers
    provider_name=AR_GOB_ENERGY_ID
    location_cache: dict[str, FuelLocation] = {}
    _url: str = AR_GOB_DATASOURCE

    async def _send_request(self, url, body) -> str:
        """Send a request to the API and return the raw response."""
        _LOGGER.debug("Sending HTTP request to %s with URL %s", self.provider_name, url)
        async with self._client_session.post(url=AR_GOB_DATASOURCE,
                                            timeout=AR_GOB_TIMEOUT,
                                            json=body) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s",
                            url,
                            response)

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites within a given radius."""
        # first query the API to populate cache / update data in case this data is unavailable.
        data = await super().search_sites(coordinates, radius)
        if len(data)>0:
            return data
        await self.update(
            areas=[{
                PROP_AREA_LAT: coordinates[0],
                PROP_AREA_LONG: coordinates[1],
                PROP_AREA_RADIUS: radius
            }],
            force=True
        )
        return await super().search_sites(coordinates, radius)

    async def update(self, areas=None, force=False) -> list[FuelLocation]:
        """Update hooks for the data source."""
        _LOGGER.debug("Starting update hook for %s to url %s", self.provider_name, self._url)
        self._configured_areas=[] if areas is None else areas
        self._clear_cache()
        if self.next_update <= datetime.now() or force:
            for area in self._configured_areas:
                try:
                    geocode = await geocode_reverse_lookup(
                        (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
                    )
                except geopyexc.GeocoderTimedOut:
                    _LOGGER.warning("Timeout occured while geocoding area %s.",
                                    area)
                    continue
                if geocode.raw["address"]["country_code"] != "ar":
                    _LOGGER.debug("Skipping area %s as not in AR.",
                                area)
                    continue
                response = json.loads(await self._send_request(
                    url=self._url,
                    body={
                        "resource_id": "80ac25de-a44a-4445-9215-090cf55cfda5",
                        "q": "",
                        "filters": {
                            "provincia": str(geocode.raw["address"]["state"]).upper(),
                            "localidad": str(geocode.raw["address"]["town"]).upper()
                        },
                        "limit": 500,
                        "offset": 0
                    }
                ))
                if response["success"]:
                    await self.parse_response(response["result"]["records"])
                else:
                    _LOGGER.error("Error sending request to %s: %s",
                                self.provider_name,
                                response)
            self.next_update = datetime.now() + self.update_interval
        return list(self.location_cache.values())

    async def parse_response(self, response) -> list[FuelLocation]:
        """Converts CMA data into fuel price mapping."""
        for location_raw in response:
            site_id = f"{self.provider_name}_{location_raw['idempresa']}"
            location = FuelLocation.create(
                    site_id=site_id,
                    name=f"{location_raw['empresabandera']} {location_raw['empresa']}",
                    address=location_raw["direccion"],
                    brand=location_raw["empresabandera"],
                    lat=location_raw["latitud"],
                    long=location_raw["longitud"],
                    available_fuels=self.parse_fuels(location_raw),
                    postal_code=None,
                    currency="ARS",
                    props={
                        PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                        PROP_FUEL_LOCATION_SOURCE_ID: location_raw["idempresa"],
                        PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True
                    },
                    next_update=self.next_update
                )
            if site_id not in self.location_cache:
                self.location_cache[site_id] = location
            else:
                await self.location_cache[site_id].update(location)
        return list(self.location_cache.values())

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Parses fuel data."""
        return [
            Fuel(
                fuel_type=fuels["producto"],
                props={
                    "fuel_id": fuels["idproducto"],
                    "valid_from": fuels["fecha_vigencia"]
                },
                cost=fuels["precio"]
            )
        ]



"""Representation of a single fuel."""

class Fuel:
    """Individual Fuel."""

    _fuel_type: str
    _cost: float
    _props: dict

    def __init__(self, fuel_type: str, cost: float, props: dict=None):
        """Initalize a fuel price."""
        self._fuel_type = fuel_type
        if cost is None:
            self._cost = 0
        else:
            self._cost = cost
        self._props = props

    @property
    def __dict__(self) -> dict:
        """Convert to a dict."""
        return {
            "type": self._fuel_type,
            "cost": self._cost,
            "props": self._props
        }

    def update(self, fuel_type: str, cost: float, props: dict):
        """Update this instance of data."""
        self._fuel_type = fuel_type
        if cost is None:
            self._cost = 0
        else:
            self._cost = cost
        self._props = props

    @property
    def fuel_type(self) -> str:
        return self._fuel_type

    @property
    def cost(self) -> float:
        return self._cost

    @property
    def props(self) -> dict | None:
        return self._props



"""Austria spritpreisrechner data source."""

from datetime import datetime

import asyncio
import logging
import json

from pyfuelprices.const import (
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_SOURCE_ID,
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS,
    DESKTOP_USER_AGENT,
)
from pyfuelprices.fuel_locations import FuelLocation, Fuel
from pyfuelprices.sources import Source
from pyfuelprices.helpers import geocode_reverse_lookup

_LOGGER = logging.getLogger(__name__)

CONST_API_BASE = "https://www.spritpreisrechner.at/api"
CONST_API_SEARCH_REGIONS = (
    f"{CONST_API_BASE}/search/gas-stations/by-region?"
    "code={REG_CODE}&fuelType={FUEL_TYPE}&includeClosed=true&type={REG_TYPE}"
)
CONST_API_SEARCH_ADDRESS = (
    f"{CONST_API_BASE}/search/gas-stations/by-address?"
    "latitude={LAT}&longitude={LONG}&fuelType={FUEL_TYPE}&includeClosed=true"
)
CONST_API_GET_REGIONS = f"{CONST_API_BASE}/regions"
CONST_FUELS = ["DIE", "SUP", "GAS"]


class SpripreisrechnerATSource(Source):
    """Data source for spritpreisrechner."""

    provider_name = "spritpreisrechner"
    _headers = {"User-Agent": DESKTOP_USER_AGENT}
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None
    _regions = []

    async def _send_request(self, url):
        """Send a HTTP request to the API."""
        async with self._client_session.get(url, headers=self._headers) as response:
            if response.ok:
                return await response.text()
            _LOGGER.error("Error sending request to %s: %s", url, response)

    async def _load_regions(self):
        """Load regions into memory."""
        self._regions = json.loads(await self._send_request(CONST_API_GET_REGIONS))

    def _parse_raw_fuel_station(self, station, site_id) -> FuelLocation:
        """Parse a raw fuel station."""
        _LOGGER.debug("Parsing %s location ID %s", self.provider_name, site_id)
        loc = FuelLocation.create(
            site_id=site_id,
            name=station.get("name", f"Unknown {station['id']}"),
            address=station["location"]["address"],
            lat=station["location"]["latitude"],
            long=station["location"]["longitude"],
            brand="unavailable",
            available_fuels=[
                Fuel(
                    fuel_type=f["label"],
                    cost=f["amount"],
                    props=f
                ) for f in station["prices"]
            ],
            postal_code=station["location"]["postalCode"],
            currency="EUR",
            props={
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station["id"],
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
                "data": station,
            },
        )
        loc.next_update = self.next_update + self.update_interval
        return loc

    def _update_fuel_station_prices(self, station, site_id):
        """Internal method to update station prices."""
        if len(station["prices"]) > 0:
            fuel = station["prices"][0]
            try:
                self.location_cache[site_id].get_fuel(fuel["label"]).update(
                    fuel_type=fuel["label"], cost=fuel["amount"], props=fuel
                )
            except ValueError:
                self.location_cache[site_id].available_fuels.append(
                    Fuel(fuel_type=fuel["label"], cost=fuel["amount"], props=fuel)
                )
            self.location_cache[site_id].last_updated = datetime.now()
            self.location_cache[site_id].next_update = self.next_update + self.update_interval

    async def update_area(self, area: dict):
        """Update given areas."""
        _LOGGER.debug("Searching FuelGR for FuelLocations at area %s", area)
        parser_coords = (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
        geocoded = await geocode_reverse_lookup(parser_coords)
        if geocoded is None:
            _LOGGER.debug("Geocode failed, skipping area %s", area)
            return
        if geocoded.raw["address"]["country_code"] not in ["at"]:
            _LOGGER.debug("Geocode not within AT, skipping area %s", area)
            return
        for fuel in CONST_FUELS:
            await self.parse_response(
                json.loads(
                    await self._send_request(
                        url=CONST_API_SEARCH_ADDRESS.format(
                            LAT=area[PROP_AREA_LAT], LONG=area[PROP_AREA_LONG], FUEL_TYPE=fuel
                        )
                    )
                )
            )

    async def update_region(self, region_code, region_type):
        """Update fuels for a given region."""
        for fuel in CONST_FUELS:
            url = CONST_API_SEARCH_REGIONS.format(
                REG_CODE=region_code, FUEL_TYPE=fuel, REG_TYPE=region_type
            )
            await self.parse_response(json.loads(await self._send_request(url)))

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites within a given radius."""
        # first query the API to populate cache / update data in case this data is unavailable.
        data = await super().search_sites(coordinates, radius)
        if len(data)>0:
            return data
        await self.update(
            areas=[{
                PROP_AREA_LAT: coordinates[0],
                PROP_AREA_LONG: coordinates[1],
                PROP_AREA_RADIUS: radius
            }],
            force=True
        )
        return await super().search_sites(coordinates, radius)

    async def update(self, areas=None, force: bool | None = None) -> list[FuelLocation]:
        """Custom update handler to query each region."""
        if (datetime.now() > self.next_update) or force:
            coros = [self.update_area(x) for x in areas]
            if len(self._regions) == 0:
                await self._load_regions()
            coros.extend(
                [self.update_region(str(region["code"]), str(region["type"]))
                 for region in self._regions]
            )
            asyncio.gather(*coros)
            return list(self.location_cache.values())

    async def parse_response(self, response) -> list[FuelLocation]:
        """Parse response into a list of stations."""
        _LOGGER.debug("Found %s fuel stations", len(response))
        for station in response:
            site_id = f"{self.provider_name}_{station['id']}"
            if site_id in self.location_cache:
                self._update_fuel_station_prices(station, site_id)
                continue
            loc = self._parse_raw_fuel_station(station, site_id)
            self.location_cache[loc.id] = loc
            continue

    def parse_fuels(self, fuels) -> list[Fuel]:
        """Method not used."""



"""GasPass data source."""

import logging
import json

from datetime import datetime, timedelta
from geopy import distance

from pyfuelprices.const import (
    PROP_AREA_LAT,
    PROP_AREA_LONG,
    PROP_AREA_RADIUS,
    PROP_FUEL_LOCATION_SOURCE,
    PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP,
    PROP_FUEL_LOCATION_SOURCE_ID,
)
from pyfuelprices.fuel_locations import Fuel, FuelLocation
from pyfuelprices.sources import (
    Source,
    geocode_reverse_lookup,
    geopyexc,
    UpdateFailedError
)

_LOGGER = logging.getLogger(__name__)

CONST_API_HOST = "https://gaspassproxy2-dhf8a6aphxbng9g4.brazilsouth-01.azurewebsites.net"
CONST_GET_FUELS = f"{CONST_API_HOST}/api/v1/gaspass/"
CONST_FUELS = [
    "ultimo_preco_alcool",
    "ultimo_preco_diesel",
    "ultimo_preco_diesel_s10",
    "ultimo_preco_gasolina",
    "ultimo_preco_gasolina_adt",
    "ultimo_preco_gnv"
]

class GasPassSource(Source):
    """GasPass Source."""
    provider_name = "gaspass"
    location_cache: dict[str, FuelLocation] = {}

    def __init__(self, update_interval = ..., client_session = None):
        if update_interval.seconds < 3600:
            update_interval = timedelta(hours=1)
        super().__init__(update_interval, client_session)

    async def _send_request(self, lat, long, radius):
        """Send a request to the API for a given postcode and radius."""
        radius = radius * 1.609344 # convert miles to km
        _LOGGER.debug("Sending request to gaspass: %s",
                      CONST_GET_FUELS)
        async with self._client_session.request(
            method="POST",
            url=CONST_GET_FUELS,
            json={
                "lat": lat,
                "long": long,
                "radius": radius
            },
            headers={"Content-Type": "application/json"}) as response:
            if not response.ok:
                text = await response.text()
                _LOGGER.error("Error sending request to %s: %s",
                                CONST_GET_FUELS,
                                text)
                raise UpdateFailedError(
                    response.status,
                    response=text,
                    headers=response.headers,
                    service=self.provider_name
                )
            return await response.text()

    async def search_sites(self, coordinates, radius: float) -> list[dict]:
        """Return all available sites within a given radius."""
        # first query the API to populate cache / update data in case this data is unavailable.
        data = await super().search_sites(coordinates, radius)
        if len(data)>0:
            return data
        await self.update(
            areas=[{
                PROP_AREA_LAT: coordinates[0],
                PROP_AREA_LONG: coordinates[1],
                PROP_AREA_RADIUS: radius
            }],
            force=True
        )
        return await super().search_sites(coordinates, radius)

    async def update(self, areas=None, force=None) -> list[FuelLocation]:
        """Custom update handler as this needs to query GasPass on areas."""
        if datetime.now() > self.next_update or force:
            self._configured_areas=[] if areas is None else areas
            for area in self._configured_areas:
                try:
                    geocode = await geocode_reverse_lookup(
                        (area[PROP_AREA_LAT], area[PROP_AREA_LONG])
                    )
                except geopyexc.GeocoderTimedOut:
                    _LOGGER.warning("Timeout occured while geocoding area %s.",
                                    area)
                    continue
                if geocode.raw["address"]["country_code"] != "br":
                    _LOGGER.debug("Skipping area %s as not in BR.",
                                area)
                    continue
                response_raw = json.loads(await self._send_request(
                    lat=area[PROP_AREA_LAT],
                    long=area[PROP_AREA_LONG],
                    radius=area[PROP_AREA_RADIUS]
                ))
                if response_raw["message"] == "ok":
                    await self.parse_response(response_raw)
                else:
                    _LOGGER.error("Error sending request to %s: %s",
                                self.provider_name,
                                response_raw)
            self.next_update += self.update_interval
        return list(self.location_cache.values())

    def _parse_raw(self, station: dict) -> FuelLocation:
        """Parse a single raw instance of a fuel site."""
        site_id = f"{self.provider_name}_{station['id']}"
        _LOGGER.debug("Parsing GasPass location ID %s", site_id)
        loc = FuelLocation.create(
            site_id=site_id,
            name=station["nome_fantasia"],
            address=station['endereco'],
            lat=station["geocode"]["latitude"],
            long=station["geocode"]["longitude"],
            brand=station["bandeira"]["nome"],
            available_fuels=self.parse_fuels(station),
            postal_code=None,
            currency="BRL",
            props={
                PROP_FUEL_LOCATION_SOURCE: self.provider_name,
                PROP_FUEL_LOCATION_SOURCE_ID: station["id"],
                PROP_FUEL_LOCATION_PREVENT_CACHE_CLEANUP: True,
            }
        )
        loc.next_update = self.next_update if self.next_update > datetime.now() else (
            self.next_update + self.update_interval
        )
        return loc

    async def parse_response(self, response) -> list[FuelLocation]:
        for station_raw in response.get("response", []):
            station = self._parse_raw(station_raw)
            if station.id not in self.location_cache:
                self.location_cache[station.id] = station
            else:
                await self.location_cache[station.id].update(station)
        return list(self.location_cache.values())

    def parse_fuels(self, fuels: dict[str, object]) -> list[Fuel]:
        """Parse fuel data from GasPass."""
        fuels_parsed = []
        for k in CONST_FUELS:
            if k in fuels and fuels[k] is not None:
                fuels_parsed.append(Fuel(
                    fuel_type=k.replace("ultimo_preco_", "").upper(),
                    cost=fuels[k],
                    props={}
                ))
        return fuels_parsed



"""Karan Retail Ltd dataprovider."""

from pyfuelprices.sources.uk import CMAParser, FuelLocation

class KaranRetailSource(CMAParser):
    """Karan Retail Ltd uses the CMA parser."""

    _url = "https://api2.krlmedia.com/integration/live_price/krl"
    provider_name = "karanretail"
    location_cache: dict[str, FuelLocation] = {}
    location_tree = None
